/* ----------------------------------------------------------------------------
 *         PalmSens EmStat SDK
 * ----------------------------------------------------------------------------
 * Copyright (c) 2016, PalmSens BV
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * PalmSens's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY PALMSENS "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL PALMSENS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */

/**
 *  Implementation of the EmStat communication protocol. See header file for more information.
 *
 *  Author(s): Hielke Veringa (hielke@palmsens.com)
 */

#include "PSComm.h"
#include "PSMethod.h" 

#include <stdio.h>

RetCode PSCommInit(PSComm* psc, DeviceType device, uint16_t read_timeout_ms,
	WriteCharFunc write_char_func, ReadCharFunc read_char_func, TimeMsFunc time_ms_func)
{
	psc->device = device;
	psc->read_timeout_ms = read_timeout_ms;
	
	psc->write_char_func = write_char_func;
	psc->read_char_func = read_char_func;
	psc->time_ms_func = time_ms_func;
	
	if(write_char_func == NULL || read_char_func == NULL || time_ms_func == NULL)
	{
		return CODE_NULL;
	}
	
	return CODE_OK;
}

//Temp buffer for values strings, int max is 10 digits, but leave space for minus and '\0'
static char ValStr[32];
static const char ParamFormatStr[] = "%s=%lu\n";

//Need to cast to unsigned long to be able to write both uint16_t and uint32_t on some devices
#define WRITE_PARAM_UINT(pPsc, pMethod, name) 								\
	sprintf(ValStr, ParamFormatStr, #name, (unsigned long)pMethod->name); 	\
	WriteStr(pPsc, ValStr);

RetCode StartMeasurement(PSComm* psc, Method* m)
{
	RetCode ret;
	WriteChar(psc, CMD_LOADMEASURE);
	ret = WaitForChar(psc, CMD_LOADMEASURE);
	
	if(ret != CODE_OK)
		return ret;
	
	WRITE_PARAM_UINT(psc, m, technique);
	WRITE_PARAM_UINT(psc, m, options);
	
	WRITE_PARAM_UINT(psc, m, Econd);
	WRITE_PARAM_UINT(psc, m, Edep);
	WRITE_PARAM_UINT(psc, m, tCond);
	WRITE_PARAM_UINT(psc, m, tDep);
	WRITE_PARAM_UINT(psc, m, tEquil);
	
	WRITE_PARAM_UINT(psc, m, cr_min);
	WRITE_PARAM_UINT(psc, m, cr_max);
	
	WRITE_PARAM_UINT(psc, m, tInt);
	
	WRITE_PARAM_UINT(psc, m, nPoints);

	//WRITE_PARAM_UINT(psc, m, limitMax);
	//WRITE_PARAM_UINT(psc, m, limitMin);
	//WRITE_PARAM_UINT(psc, m, crLimitMin);
	//WRITE_PARAM_UINT(psc, m, crLimitMax);
	if((m->technique == TECHNIQUE_DPV) || (m->technique == TECHNIQUE_NPV) || (m->technique == TECHNIQUE_PAD) || (m->technique == TECHNIQUE_SWV))
	{
		WRITE_PARAM_UINT(psc, m, tPulse);
		WRITE_PARAM_UINT(psc, m, Epulse);
	}
	
	WRITE_PARAM_UINT(psc, m, Ebegin);
	
	if((m->technique == TECHNIQUE_LSV) || (m->technique == TECHNIQUE_DPV) || (m->technique == TECHNIQUE_SWV) 
		|| (m->technique == TECHNIQUE_NPV)  || (m->technique == TECHNIQUE_CV) )
	{
		WRITE_PARAM_UINT(psc, m, Estep);
	}
	
	if((m->technique == TECHNIQUE_LSV) || (m->technique == TECHNIQUE_DPV) || (m->technique == TECHNIQUE_SWV) 
		|| (m->technique == TECHNIQUE_NPV))
	{
		WRITE_PARAM_UINT(psc, m, Eend);
	}
	
	if(m->technique == TECHNIQUE_CV)
	{
		WRITE_PARAM_UINT(psc, m, Evtx1);
		WRITE_PARAM_UINT(psc, m, Evtx2);
		WRITE_PARAM_UINT(psc, m, nScans);
	}
	
	if(m->technique == TECHNIQUE_MPAD)
	{
		WRITE_PARAM_UINT(psc, m, E1);
		WRITE_PARAM_UINT(psc, m, E2);
		WRITE_PARAM_UINT(psc, m, E3);
		
		WRITE_PARAM_UINT(psc, m, t2);
		WRITE_PARAM_UINT(psc, m, t3);
	}
	
	//WRITE_PARAM_UINT(psc, m, current);
	
	WRITE_PARAM_UINT(psc, m, cr);
	//WRITE_PARAM_UINT(psc, m, tMax);
	
	if(m->options & OPTIONS_CELL_ON_AFTER_MEASUREMENT)
		WRITE_PARAM_UINT(psc, m, Estby);
	
	WRITE_PARAM_UINT(psc, m, nmux);
	if(m->nmux > 0)
		WRITE_PARAM_UINT(psc, m, mux_delay);
	
	if(m->options & OPTIONS_MEASURE_AUX_INPUT)
		WRITE_PARAM_UINT(psc, m, ChAux);
	
	WRITE_PARAM_UINT(psc, m, nadmean);
	WRITE_PARAM_UINT(psc, m, d1);
	WRITE_PARAM_UINT(psc, m, d16);
	
	WriteChar(psc, CMD_END_OF_METHOD);
	
	//Filter out crc bytes
	char buf[5];
	ReadBuf(psc, buf, 5);
	
	return CODE_OK;
}

void StartLoadedMeasurement(PSComm* psc)
{
	WriteChar(psc, CMD_MEASURE);
}

void AbortMeasurement(PSComm* psc)
{
	WriteChar(psc, CMD_ABORT);
}

void TogglePauseMeasurement(PSComm* psc)
{
	WriteChar(psc, CMD_HOLD);
}

void ReverseDirection(PSComm* psc)
{
	WriteChar(psc, CMD_REVERSE);
}

RetCode QueryVersion(PSComm* psc, char* version_str, size_t size)
{
	RetCode ret;
	WriteChar(psc, CMD_VERSION);
	ret = WaitForChar(psc, CMD_VERSION);
	if(ret != CODE_OK)
		return ret;
	//make sure string is null terminated.
	version_str[size-1] = 0; 
	return ReadBuf(psc, version_str, size-1);
}

void Calibrate(PSComm* psc)
{
	WriteChar(psc, CMD_CALIBRATE);
}

void FastMode(PSComm* psc)
{
	WriteChar(psc, CMD_FAST_CONTROL);
}

void NormalMode(PSComm* psc)
{
	WriteChar(psc, CMD_NORMAL_CONTROL);
}

RetCode ReceivePackage(PSComm* psc, Method* m, MeasureData* ret_data)
{
	char c;		
	RetCode ret = ReadChar(psc, &c);
	if(ret != CODE_OK)
	{
		return ret;
	}

	switch(c)
	{
	case REPLY_STD_DP:
		return ReceiveTPackage(psc, ret_data);
	case REPLY_MEASURE_DP:
		return ReceiveUPackage(psc, m, ret_data);
	case REPLY_ENDOFCURVE:
		psc->current_phase = MEASURE_PHASE_NONE;
		return CODE_MEASUREMENT_DONE;
	case REPLY_MEASURING:
		return CODE_MEASURING;
	case REPLY_MUX_DP:
		return CODE_NOT_IMPLEMENTED;
	default:
		return CODE_UNEXPECTED_DATA;
	}
}

RetCode MCSetBaud(PSComm* psc, BaudEnum baud)
{
	return ManualControlCmd(psc, CMD_SETBAUDRATE, baud, 0xFF);
}

RetCode MCSetPotential(PSComm* psc, double potential)
{
	uint16_t intInt = EncodePotential(psc->device, potential);
	uint8_t intH = (intInt >> 8) & 0xFF; 	//calculate high byte as integer
	uint8_t intL = intInt & 0xFF; 			//calculate low byte as integer
	return ManualControlCmd(psc, MCMD_POTENTIAL, intL, intH);
}
/**
 * Set baud.
 */
RetCode MCSetCell(PSComm* psc, CurrentRange current_range, CellOnOff cell_on_off)
{
	return ManualControlCmd(psc, MCMD_MODE, current_range, cell_on_off);
}

/**
 * Read current, potential or auxiliary analog input on CON2
 */
RetCode MCReadChannel(PSComm* psc, Channel chan, double* ret_val)
{
	char read_buf[MANUAL_CMD_LENGTH];
	RetCode ret = ManualControlCmd(psc, MCMD_READCHANNEL, chan, 0xFF);
	if(ret != CODE_OK)
		return ret;
	
	ret = ReadBuf(psc, read_buf, MANUAL_CMD_LENGTH);
	if(ret != CODE_OK)
		return ret;
	
	if(read_buf[0] != MCMD_READCHANNEL)
		return CODE_UNEXPECTED_DATA;
	
	uint8_t intL = FromHex(read_buf+1);
	uint8_t intH = FromHex(read_buf+3);
	double vin = (intH * 256 + intL) / (double)0xFFFF * 4.096;
	
	switch(chan)
	{
	case CHAN_CURRENT:
		vin -= 2.048;
		break;
	case CHAN_POTENTIAL:
		vin = (vin - 2.048) * EmstatVals[psc->device][EFACTOR];
		break;
	case CHAN_AUX_INPUT:
		//vin = vin;
		break;
	}
	
	*ret_val = vin;
	return CODE_OK;
}

RetCode MCSetDAC1(PSComm* psc, double v_applied)
{
	int32_t v_int = v_applied * 16000;
	uint8_t intH = (uint8_t)(v_int / 256);
	uint8_t intL = (uint8_t)(v_int - 256 * intH);
	return ManualControlCmd(psc, MCMD_DAC1, intL, intH);
}

RetCode MCSetDigital(PSComm* psc, uint8_t bits)
{
	return ManualControlCmd(psc, MCMD_SETDIGITAL, bits, 0xFF);
}

RetCode MCReadDigital(PSComm* psc, uint8_t* ret_bits)
{
	char read_buf[MANUAL_CMD_LENGTH];
	RetCode ret = ManualControlCmd(psc, MCMD_READDIGITAL, 0xFF, 0xFF);
	if(ret != CODE_OK)
	{
		return ret;
	}
	ret = ReadBuf(psc, read_buf, MANUAL_CMD_LENGTH);
	if(ret != CODE_OK)
	{
		return ret;
	}
	if(read_buf[0] != MCMD_READDIGITAL)
	{
		return CODE_UNEXPECTED_DATA;
	}
	
	*ret_bits = FromHex(read_buf+1) & 0x01;
	
	return CODE_OK;
}


RetCode ReadBuf(PSComm* psc, char* buf, size_t size)
{
	uint16_t cnt = 0;
	unsigned long end_time = psc->time_ms_func() + psc->read_timeout_ms;
	while(cnt < size && psc->time_ms_func() < end_time)
	{
		int c = psc->read_char_func();
		if(c >= 0)
		{
			(*buf) = c;
			cnt++;
			buf++;
		}
	}
	if(cnt < size)
	{
		return CODE_TIMEOUT;
	}
	return CODE_OK;
}

RetCode ReadChar(PSComm* psc, char* c)
{
	unsigned long end_time = psc->time_ms_func() + psc->read_timeout_ms;
	int c_temp;
	while(psc->time_ms_func() < end_time)
	{
		c_temp = psc->read_char_func();
		if(c_temp >= 0)
		{
			*c = (char)c_temp;
			return CODE_OK;
		}
	}
	return CODE_TIMEOUT;
}

///
/// Reads from the device until 'c' is received or the read timeout has passed
/// Returns: CODE_OK if successful, otherwise CODE_TIMEOUT.
///
RetCode WaitForChar(PSComm* psc, char c)
{
	unsigned long end_time = psc->time_ms_func() + psc->read_timeout_ms;
	while(psc->time_ms_func() < end_time)
	{
		if(psc->read_char_func() == c)
		{
			return CODE_OK;
		}
	}
	return CODE_TIMEOUT;
}

void WriteChar(PSComm* psc, char c)
{
	psc->write_char_func(c);
}

void WriteHex(PSComm* psc, char data)
{
	char buf[2];
	ToHex(data, buf);
	WriteChar(psc, buf[0]);
	WriteChar(psc, buf[1]);
}

void WriteBuf(PSComm* psc, const char* buf, size_t size)
{
	size_t i;
	for(i = 0; i < size; i++)
	{
		WriteChar(psc, buf[i]);
	}
}

void WriteStr(PSComm* psc, const char* buf)
{
	while(*buf != 0)
	{
		WriteChar(psc, *buf);
		buf++;
	}
}

RetCode ManualControlCmd(PSComm* psc, char cmd, uint8_t data1, uint8_t data2)
{
	WriteChar(psc, CMD_MANUAL);

	char c;
	RetCode ret = WaitForChar(psc, CMD_MANUAL);
	if(ret != CODE_OK)
	{
		return ret;
	}
	
	WriteChar(psc, cmd);
	WriteHex(psc, data1);
	WriteHex(psc, data2);
	
	return CODE_OK;
}

RetCode ReceiveTPackage(PSComm* psc, MeasureData* ret_data)
{
	char buf[T_PACKAGE_LENGTH];
	RetCode ret = ReadBuf(psc, buf, T_PACKAGE_LENGTH);
	if(ret != CODE_OK)
		return ret;
	
	uint8_t tempByte;
	
	//We want to know the current range for decoding the current, so take this out first
	tempByte = FromHex(&buf[10]);
	ret_data->crange = (CurrentRange)(tempByte & 0x0F);
	ret_data->status = (Status)(tempByte & 0xF0);
	
	ret_data->potential = DecodePotential(psc->device, FromHex(&buf[0]), FromHex(&buf[2]), EFACTOR);

	ret_data->current = DecodeCurrent(psc->device, FromHex(&buf[4]), FromHex(&buf[6]), 0, ret_data->crange, EFACTOR);
	
	ret_data->phase = (MeasurePhase) FromHex(&buf[8]);
	psc->current_phase = ret_data->phase;
	
	ret_data->aux_input = DecodeAuxPotential(psc->device, FromHex(&buf[12]), FromHex(&buf[14]));
	
	ret_data->noise = DecodeNoise(psc->device, FromHex(&buf[16]), FromHex(&buf[18]));
	
	return CODE_OK;
}

RetCode ReceiveUPackage(PSComm* psc, Method* m, MeasureData* ret_data)
{
	char buf[U_PACKAGE_LENGTH];
	RetCode ret = ReadBuf(psc, buf, U_PACKAGE_LENGTH);
	if(ret != CODE_OK)
		return ret;
	
	uint8_t tempByte;
	
	//We want to know the current range for decoding the current, so take this out first
	tempByte = FromHex(&buf[10]);
	ret_data->crange = (CurrentRange)(tempByte & 0x0F);
	ret_data->status = (Status)(tempByte & 0xF0);
	
	if(m->technique == TECHNIQUE_CP) //OCP measurements have a different data order
	{
		ret_data->potential = DecodePotential(psc->device, FromHex(&buf[4]), FromHex(&buf[6]), EFACTOR);
		ret_data->current = DecodeCurrent(psc->device, FromHex(&buf[0]), FromHex(&buf[2]), FromHex(&buf[8]), ret_data->crange, DACFACTOR);
	}
	else
	{
		ret_data->potential = DecodePotential(psc->device, FromHex(&buf[0]), FromHex(&buf[2]), DACFACTOR);
		ret_data->current = DecodeCurrent(psc->device, FromHex(&buf[4]), FromHex(&buf[6]), FromHex(&buf[8]), ret_data->crange, EFACTOR);
	}
	ret_data->phase = MEASURE_PHASE_MEASURING;
	psc->current_phase = MEASURE_PHASE_MEASURING;
	
	ret_data->aux_input = DecodeAuxPotential(psc->device, FromHex(&buf[12]), FromHex(&buf[14]));
	
	ret_data->noise = 0; //Not measured
	
	return CODE_OK;
}

