/* ----------------------------------------------------------------------------
 *         PalmSens EmStat SDK
 * ----------------------------------------------------------------------------
 * Copyright (c) 2016, PalmSens BV
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * PalmSens's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY PALMSENS "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL PALMSENS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */

 /**
 *  PSComm handles the communication with the EmStat. 
 *	Only this file needs to be included in your software.
 *	To communicate with an EmStat, create a PSComm struct and call PSCommInit() on it.
 *	If communication with multiple EmStats is required, create a separate PSComm with different
 * 	communication ports (defined by the write_char_func and read_char_func) for each EmStat.
 *
 *	Once a PSComm struct has been successfully created, create a Method struct and call one of the
 *	SetParams functions in PSMethod.h to "fill" it for a specific measurement type. 
 *	StartMeasurement can then be called to start a measurement, passing the filled method struct.
 *	
 *	To receive data from the EmStat, call ReceivePackage() periodically. 
 *	By default, the EmStat sends packages both when idle and when measuring.
 *
 *	Example:
 *
 *	int main()
 *	{
 *		Method m;
 *		PSComm psc;
 *		PSCommInit(&psc, EMSTAT3, 1000, &write_func, &read_func, &time_ms_func);
 *
 *		SetLinearSweepParams(&m, psc.device,
 *      	0, 4.0, 1.5, 4.0, 
 *      	6.0, CR_10nA, CR_1mA, 
 *         	-2.0, 2.0, 0.1, 0.1, 
 *         	OPTIONS_NONE, 0.0, 50);
 *
 *		StartMeasurement(&psc, &m);
 *		
 *		MeasureData data;
 *		while(ReceivePackage(&psc, &m, &data) != CODE_MEASUREMENT_DONE)
 *		{
 *			printf("Potential: %f Current: %f", data.potential, data.current);
 *		}
 *
 *		return 0;
 *	}
 *
 *  Author(s): Hielke Veringa (hielke@palmsens.com)
 * 	Version: 1.23
 */
 
#ifndef PSCOMM_H
#define PSCOMM_H

//////////////////////////////////////////////////////////////////////////////
// Includes
//////////////////////////////////////////////////////////////////////////////

#include <stdint.h>
#include <stddef.h>
#include "PSMethod.h"
#include "PSCommon.h"

//////////////////////////////////////////////////////////////////////////////
// Constants
//////////////////////////////////////////////////////////////////////////////
#define VERSION_STR_LENGTH	(8 + 1)
#define MANUAL_CMD_LENGTH	5

#define U_PACKAGE_LENGTH 	16
#define T_PACKAGE_LENGTH 	20

//////////////////////////////////////////////////////////////////////////////
// Types
//////////////////////////////////////////////////////////////////////////////

///
/// Whether the cell should be on or off.
///
typedef enum _CellOnOff
{
	CELL_ON	 = 0x03,
	CELL_OFF = 0x05,
} CellOnOff;

///
/// Current status, if there is a underload or overload that datapoint may be inaccurate.
///
typedef enum _Status
{
	///Current overload
	STATUS_OVERLOAD = 0x20,
	///Current underload
	STATUS_UNDERLOAD = 0x40,
} Status;

///
/// Available BAUD settings for EmStat, default is BAUD_230400
///
typedef enum _BaudEnum
{
	BAUD_9600 	= 0x05,
	BAUD_19200 	= 0x06,
	BAUD_38400	= 0x07,
	BAUD_57600	= 0x08,
	BAUD_115200	= 0x09,
	BAUD_230400 = 0x0A,
} BaudEnum;

///
/// Measurement phase the device is in, updated only when receiving 'U' or 'T' packages.
///
typedef enum _MeasurePhase 
{ 
	MEASURE_PHASE_NONE = 0,
	MEASURE_PHASE_CONDITIONING = 1,
	MEASURE_PHASE_DEPOSITIONING = 2,
	MEASURE_PHASE_EQUILIBRATING = 3,
	MEASURE_PHASE_MEASURING = 4,
	MEASURE_PHASE_WAITING = 5 //used in between measurements in mixed mode
} MeasurePhase;

///
/// Available commands on the EmStat.
///
typedef enum _Command
{
  	CMD_VERSION 		= 't',
  	CMD_LOADMEASURE 	= 'L',
	CMD_END_OF_METHOD	= '*',
  	CMD_MEASURE 		= 'M',
	CMD_MEASURE_CONT	= 'C',
	CMD_SKIP_PRETREAT	= 'Y',
	CMD_DISPLAY_HEX		= 'k',
	CMD_GET_METHOD		= 'K',
	CMD_CALIBRATE		= 'N',
	CMD_HOLD			= 'z',
	CMD_REVERSE			= 'R',
  	CMD_ABORT 			= 'Z',
  	CMD_MANUAL 			= 'c',
	CMD_DOWNLOAD		= 'I',
	CMD_FAST_CONTROL 	= 'J',
	CMD_NORMAL_CONTROL	= 'j',
	CMD_SETBAUDRATE		= 'B',
} Command;

///
/// Available manual commands on the EmStat.
///
typedef enum _MCommand
{
  	MCMD_POTENTIAL 		= 'D',
	MCMD_CAPABILITIES	= 'C',
  	MCMD_MODE 			= 'G',
	MCMD_BIPOT_MODE		= 'W',
	//MCMD_BIPOT_POTENTIAL	= 'd',
	MCMD_DAC1			= 'd',
  	MCMD_READCHANNEL 	= 'a',
	//MCMD_READCHANNEL_INTERVAL='b',
  	MCMD_READDIGITAL 	= 'r',
  	MCMD_SETDIGITAL 	= 'v',
	MCMD_READ_EEPROM	= 'h',
	MCMD_WRITE_EEPROM 	= 'H',
	MCMD_DAC_OFFSET		= 'O',
	MCMD_E_OFFSET		= 'o',
} MCommand;

///
/// Possible replies from the EmStat.
///
typedef enum _Reply
{
	//REPLY_WAITING		= 'X',	 	//OBSOLETE
	//REPLY_IDLE			= 'Q', 	//OBSOLETE
	REPLY_STD_DP		= 'T',
	REPLY_MEASURE_DP	= 'U',
	REPLY_MUX_DP		= 'P',
	//REPLY_AMPDET_CURVE	= 'V', 	//OBSOLETE
	//REPLY_AMPDET_DP		= 'Y', 	//OBSOLETE
	//REPLY_CHRONOPOT_DP	= 'Y', 	//OBSOLETE
	//REPLY_BUFFER		= 'S',		//OBSOLETE
	//REPLY_CHECKSUM		= 'k',	//OBSOLETE
	REPLY_MEASURING		= 'M',
	REPLY_ENDOFCURVE	= '*',
	REPLY_UNKNOWNCMD	= '?',
	REPLY_ERROR			= '!',
	REPLY_STRING		= '#'
} Reply;

///
/// The communication object for one EmStat.
/// You can instantiate multiple PSComms if you have multiple EmStats,
/// but you will need write / read functions from separate (serial) ports to talk to them.
///
typedef struct _PSComm
{
	WriteCharFunc write_char_func;
	ReadCharFunc read_char_func;
	TimeMsFunc time_ms_func;
	uint16_t read_timeout_ms;
	DeviceType device;
	MeasurePhase current_phase;
} PSComm;

///
/// Encapsulates the data packages received from the emstat,
/// for both 'U'(Measurement) and 'T' (Idle) packages.
///
typedef struct _MeasureData
{
	/// Potential in Volts
	float potential;
	/// Current in Current Range
	float current;
	/// Aux input, unit depends on type of input
	float aux_input;
	/// Current range the current is expressed in.
	CurrentRange crange;
	/// The phase of the measurement.
	MeasurePhase phase;
	/// The overload status of the datapoint.
	Status status;
	/// A measure of external noise on the WE, only measured in T packages.
	float noise;
} MeasureData;

//////////////////////////////////////////////////////////////////////////////
// Normal Functions
//
// These functions are used during normal operation.
//////////////////////////////////////////////////////////////////////////////

///
/// Initialize the PSComm object.
///
/// psc:				The PSComm data struct.
/// device: 			The device type you wish to connect to.
/// read_timeout_ms: 	Timeout for all reads performed by PSComm.
/// write_char_func: 	Function pointer to the write function this PSComm should use.
/// read_char_func: 	Function pointer to the read function this PSComm should use.
/// time_ms_func:		Function pointer to a function that returns a time in ms.
///
/// Returns: CODE_OK if successful, otherwise CODE_NULL.
///
RetCode PSCommInit(PSComm* psc, DeviceType device, uint16_t read_timeout_ms,
	WriteCharFunc write_char_func, ReadCharFunc read_char_func, TimeMsFunc time_ms_func);
	
/// 
/// Sends method to device and starts the measurement. If using a multiplexer, 
/// MCSetDigital() can be used to change the multiplexer channel that's being measured.
/// 
/// psc:	The PSComm data struct.
/// m:	 	The to start.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA or CODE_TIMEOUT.
///
RetCode StartMeasurement(PSComm* psc, Method* m);

///
/// Start measurement using the loaded method, stored in memory.
/// Use "StartMeasurement" if you want to start a different measurement.
///
/// psc:	The PSComm data struct.
///
void StartLoadedMeasurement(PSComm* psc);

///
/// Abort current measurement.
///
/// psc:	The PSComm data struct.
///
void AbortMeasurement(PSComm* psc);

///
/// Pause LSV, CV or AD or PAD measurement, call TogglePauseMeasurement again to resume.
///
/// psc:	The PSComm data struct.
///
void TogglePauseMeasurement(PSComm* psc);

///
/// Reverse scan direction during CV measurement.
///
/// psc:	The PSComm data struct.
///
void ReverseDirection(PSComm* psc);

///
/// Receive hardware version, software (firmware) version.
/// Standard version format (## represents the firmware version):
/// EmStat1-2	: "EMSTAT##"  (if version is <6.2, model is EmStat1, otherwise EmStat2)
/// EmStat3		: "EMST3 ##"
/// EmStat3+	: "EMST3P##"
/// 
/// psc:			The PSComm data struct.
/// version_str: 	String buffer that will be filled with the version string, 
///					must have "size" allocated space.
/// size: 			Expected size of version string, typically VERSION_STR_LENGTH.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA or CODE_TIMEOUT.
///
RetCode QueryVersion(PSComm* psc, char version_str[], size_t size);

///
/// Calibrate analog electronics.
///
/// psc:	The PSComm data struct.
///
void Calibrate(PSComm* psc);

///
/// Set EmStat in the ‘Fast Response’ mode. In this mode no idle packages containing 
/// the EmStat's status will be sent.
///
/// psc:	The PSComm data struct.
///
void FastMode(PSComm* psc);

///
/// Set EmStat in the ‘Normal Response’ mode. In this mode idle packages containing 
/// the EmStat's status will be sent. This is the default mode.
/// Use ReceivePackage() to receive and parse these packages.
///
/// psc:	The PSComm data struct.
///
void NormalMode(PSComm* psc);

///
/// Wait for a package or until the timeout has passed and parse it.
/// Currents are expressed in the Current Range, potentials are expressed in Volts.
///
/// psc:		The PSComm data struct.
/// method:		The Method used to measure this package, if any (can be null).
/// ret_data: 	The package received is parsed and put into this struct.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA, CODE_TIMEOUT 
/// 		 or CODE_NOT_IMPLEMENTED.
///
RetCode ReceivePackage(PSComm* psc, Method* m, MeasureData* ret_data);

//////////////////////////////////////////////////////////////////////////////
// Manual Control Functions
//
// These commands can be used if the EmStat's built-in measurements don't do exactly
// what is required. They are meant for somewhat more advanced usage.
//////////////////////////////////////////////////////////////////////////////

///
/// Set EmStat BAUD (Communication speed). The set BAUD is saved on the device and will
/// remain even after resetting the device.
///
/// psc:	The PSComm data struct.
/// baud:	The BAUD to set (BAUD is equal to bits/s in this case)
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA, CODE_TIMEOUT.
///
RetCode MCSetBaud(PSComm* psc, BaudEnum baud);

///
/// Manually set the output potential of the EmStat.
/// This potential is only actually output when the cell is on.
///
/// psc:		The PSComm data struct.
/// potential:	The output potential in Volts.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA, CODE_TIMEOUT.
///
RetCode MCSetPotential(PSComm* psc, double potential);

///
/// Manually set cell state (the cell is used to output a potential during measurements).
///
/// psc:			The PSComm data struct.
/// current_range:	The current EmStats current range, set this to just above the max current 
///					you are expecting.
/// cell_on_off:	Turns the cell on or off. If the cell is off, no potential is output.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA, CODE_TIMEOUT.
///
RetCode MCSetCell(PSComm* psc, CurrentRange current_range, CellOnOff cell_on_off);

///
/// Manually read the current or potential from the selected channel on CON2.
///
/// psc:		The PSComm data struct.
/// chan:		The channel to read from (current, potential or auxiliary input).
/// ret_val:	Returns the current or potential read from the channel.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA, CODE_TIMEOUT.
///
RetCode MCReadChannel(PSComm* psc, Channel chan, double* ret_val);

///
/// Sets the analog output level on pin 10 of CON2. 
///
/// psc:		The PSComm data struct.
/// v_applied:	The analog output level in Volts.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA, CODE_TIMEOUT.
///
RetCode MCSetAnalog(PSComm* psc, double v_applied);

///
/// Sets the digital outputs (CON1 pins 1,2,3 and 4) on the EmStat, 
/// also used to change the multiplexer channel.
///
/// psc:	The PSComm data struct.
/// bits:	If a multiplexer is connected, input the channel index here to change the channel.
///			If no multiplexer is connected, input a bitmask here to set the output pins to.
///			Only values of 0x00 to 0x0F (0 to 15) are valid.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA, CODE_TIMEOUT.
///
RetCode MCSetDigital(PSComm* psc, uint8_t bits);

///
/// Reads the digital input on the EmStat (CON2 pin 5).
///
/// psc:		The PSComm data struct.
/// ret_bits:	Returns the status of pin 5 on CON2. 1 if high, 0 if low.
///
/// Returns: CODE_OK if successful, otherwise CODE_UNEXPECTED_DATA, CODE_TIMEOUT.
///
RetCode MCReadDigital(PSComm* psc, uint8_t* ret_bits);


//////////////////////////////////////////////////////////////////////////////
// Internal Communication Functions
//
// These functions are used by the SDK internally to communicate with the emstat.
// You probably don't need to use these directly, but they're here so you can if you want.
//////////////////////////////////////////////////////////////////////////////


///
/// Reads a character buffer using the supplied read_char_func
/// Returns: CODE_OK if successful, otherwise CODE_TIMEOUT.
///
RetCode ReadBuf(PSComm* psc, char* buf, size_t size);

///
/// Reads a character using the supplied read_char_func
/// Returns: CODE_OK if successful, otherwise CODE_TIMEOUT.
///
RetCode ReadChar(PSComm* psc, char* c);

///
/// Reads from the device until 'c' is received or the read timeout has passed
/// Returns: CODE_OK if successful, otherwise CODE_TIMEOUT.
///
RetCode WaitForChar(PSComm* psc, char c);

///
/// Writes a character using the supplied write_char_func
///
void WriteChar(PSComm* psc, char c);

///
/// Writes "data" as a hex string using the supplied write_char_func
///
void WriteHex(PSComm* psc, char data);

///
/// Writes a character buffer using the supplied write_char_func
///
void WriteBuf(PSComm* psc, const char* buf, size_t size);

///
/// Writes a 0 terminated string using the supplied write_char_func
///
void WriteStr(PSComm* psc, const char* buf);

///
/// Sends a manual command to the EmStat.
///
RetCode ManualControlCmd(PSComm* psc, char cmd, uint8_t data1, uint8_t data2);

///
/// Receive and parse idle package
/// Currents are expressed in the Current Range, potentials are expressed in Volts.
///
RetCode ReceiveTPackage(PSComm* psc, MeasureData* ret_data);

///
/// Receive and parse measurement package
/// Currents are expressed in the Current Range, potentials are expressed in Volts.
///
RetCode ReceiveUPackage(PSComm* psc, Method* m, MeasureData* ret_data);

#endif //PSCOMM_H
