/* ----------------------------------------------------------------------------
 *         PalmSens EmStat SDK
 * ----------------------------------------------------------------------------
 * Copyright (c) 2016, PalmSens BV
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * PalmSens's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY PALMSENS "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL PALMSENS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */
 
 /**
 *  Implementation of the EmStat method data structure. See header file for more information.
 *
 *  Author(s): Hielke Veringa (hielke@palmsens.com)
 */
 
#include "PSMethod.h"

#include <math.h>
#include <string.h>

const double EmstatVals[NUM_DEVICES][NUM_DEVICE_PARAMS] = {
//					Efactor		DACfactor 	Vrange
/* EmStat2  */ {	1, 			1, 			2,		},
/* EmStat3  */ {	1.5, 		1.599, 		3,		},
/* EmStat3+ */ {	2, 			2, 			4,		},
};

void SetLinearSweepParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Eend, double Estep, double scan_rate, 
	Options options, double Estby, uint8_t power_freq)
{
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_LSV;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = tEquil;
	
	m->cr_min = (uint8_t)cr_min;
	m->cr_max = (uint8_t)cr_max;
	m->cr = (uint8_t)cr_max;
	
	m->Ebegin = EncodePotential(device, Ebegin);
	m->Estep = EncodePotentialStep(device, Estep);
	m->nPoints = round(fabs(Ebegin - Eend) / fabs(Estep)) + 1;
	
	m->options = (uint16_t) options;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	double interval = fabs(Estep) / scan_rate;
	SetInterval(m, device, interval);
	SetADTiming(m, device, interval / 4, power_freq);
}


void SetDifferentialPulseParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Eend, double Estep, double Epulse, double tPulse, double scan_rate, 
	Options options, double Estby, uint8_t power_freq)
{
	double tMeas;
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_DPV;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = tEquil;
	
	m->cr_min = (uint8_t)cr_min;
	m->cr_max = (uint8_t)cr_max;
	m->cr = (uint8_t)cr_max;
	
	m->Ebegin = EncodePotential(device, Ebegin);
	m->Estep = EncodePotentialStep(device, Estep);
	m->nPoints = round(fabs(Ebegin - Eend) / fabs(Estep)) + 1;
	
	m->options = (uint16_t) options;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	SetInterval(m, device, fabs(Estep) / scan_rate);
	tMeas = SetADTiming(m, device, tPulse / 4, power_freq);
	
	m->Epulse = EncodePotentialStep(device, Epulse);
	m->tPulse = round((tPulse - tMeas)/0.0000152);
}

void SetSquareWaveParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Eend, double Estep, double amplitude, double frequency,
	Options options, double Estby, uint8_t power_freq)
{
	double tMeas;
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_SWV;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = tEquil;
	
	m->cr_min = (uint8_t)cr_min;
	m->cr_max = (uint8_t)cr_max;
	m->cr = (uint8_t)cr_max;
	
	m->Ebegin = EncodePotential(device, Ebegin);
	m->Estep = EncodePotentialStep(device, Estep);
	m->nPoints = round(fabs(Ebegin - Eend) / fabs(Estep)) + 1;
	
	m->options = (uint16_t) options;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	SetInterval(m, device, 1.0 / frequency );
	tMeas = SetADTiming(m, device, 1.0 / (6.0 * frequency), power_freq);
	
	m->Epulse = EncodePotentialStep(device, amplitude);
	m->tPulse = round((0.5 / frequency - tMeas) / 15.2e-6);
}

void SetNormalPulseParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Eend, double Estep, double tPulse, double scan_rate,
	Options options, double Estby, uint8_t power_freq)
{
	double tMeas;
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_NPV;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = tEquil;
	
	m->cr_min = (uint8_t)cr_min;
	m->cr_max = (uint8_t)cr_max;
	m->cr = (uint8_t)cr_max;
	
	m->Ebegin = EncodePotential(device, Ebegin);
	m->Estep = EncodePotentialStep(device, Estep);
	m->nPoints = round(fabs(Ebegin - Eend) / fabs(Estep)) + 1;
	
	m->options = (uint16_t) options;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	SetInterval(m, device, fabs(Estep) / scan_rate );
	tMeas = SetADTiming(m, device, tPulse / 4, power_freq);
	
	m->tPulse = round((tPulse - tMeas)/0.0000152);
}

void SetCyclicVoltammetryParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Evtx1, double Evtx2, double Estep, double scan_rate, uint8_t nScans,
	Options options, double Estby, uint8_t power_freq)
{
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_CV;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = tEquil;
	
	m->cr_min = (uint8_t)cr_min;
	m->cr_max = (uint8_t)cr_max;
	m->cr = (uint8_t)cr_max;
	
	m->Ebegin = EncodePotential(device, Ebegin);
	m->Evtx1 = EncodePotential(device, Evtx1);
	m->Evtx2 = EncodePotential(device, Evtx2);
	
	int idir = (fabs(Ebegin - Evtx1) < Estep) ? SIGN(Evtx2 - Evtx1) : SIGN(Evtx1 - Evtx2);
	
	m->Estep = EncodePotentialStep(device, Estep * idir);
	
	m->options = (uint16_t) options;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	m->nScans = nScans;
	
	double tInt = fabs(Estep) / scan_rate;
	SetInterval(m, device, tInt);
	SetADTiming(m, device, tInt / 4, power_freq);
}

void SetAmperometricDetectionParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Edc, double tInterval, double tRun,
	Options options, double Estby, uint8_t power_freq)
{
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_AD;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = tEquil;
	
	m->cr_min = (uint8_t)cr_min;
	m->cr_max = (uint8_t)cr_max;
	m->cr = (uint8_t)cr_max;
	
	m->Ebegin = EncodePotential(device, Edc);
	m->nPoints = floor((tRun / tInterval) + 1);
	
	m->options = (uint16_t) options;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	SetInterval(m, device, tInterval);
	uint8_t divisor = (options & OPTIONS_MEASURE_AUX_INPUT) ? 3 : 2;
	SetADTiming(m, device, tInterval / divisor, power_freq);
}

void SetPulsedAmperometricDetectionParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Edc, double Epulse, double tPulse, PADMode mode, double tInterval, double tRun,
	Options options, double Estby, uint8_t power_freq)
{
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_PAD;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = tEquil;
	
	m->cr_min = (uint8_t)cr_min;
	m->cr_max = (uint8_t)cr_max;
	m->cr = (uint8_t)cr_max;
	
	m->Ebegin = EncodePotential(device, Edc);
	m->Epulse = EncodePotential(device, Epulse);
	m->nPoints = floor((tRun / tInterval) + 1);
	
	m->options = (uint16_t) options | (uint16_t) mode;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	double tMeas;
	switch (mode)
	{
		case PAD_MODES_DC:
			tMeas = (tInterval - tPulse) / 2.0;
			break;
		case PAD_MODES_PULSE:
			tMeas = tPulse / 2.0;
			break;
		case PAD_MODES_DIFFERENTIAL:
			if ((tInterval - tPulse) < tPulse)
				tMeas = (tInterval - tPulse) / 2.0;
			else
				tMeas = tPulse / 2.0;
			break;
	}
	tMeas = SetADTiming(m, device, tMeas, power_freq);
	
	switch (mode)
	{
		case PAD_MODES_DC:
			m->tPulse = round(tPulse / 71.5e-6); // twaitlong is used
			break;
		case PAD_MODES_PULSE:
		case PAD_MODES_DIFFERENTIAL:
			m->tPulse = round((tPulse - tMeas) / 71.5e-6);
			break;
	}
				
	SetInterval(m, device, tInterval);
}

void SetOpenCircuitPotentiometryParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep,
	double tInterval, double tRun,
	Options options, double Estby, uint8_t power_freq)
{
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_CP;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = 0;
	
	m->Ebegin = EncodePotential(device, 0.0); //Probably not required
	
	m->nPoints = floor((tRun / tInterval) + 1);
	
	m->options = (uint16_t) options;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	SetInterval(m, device, tInterval);
	SetADTiming(m, device, tInterval / 2, power_freq);
}

void SetMultiplePulseAmperometryParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double E1, double t1, double E2, double t2, double E3, double t3, double tRun,
	Options options, double Estby, uint8_t power_freq)
{
	//Clear method
	memset(m, 0, sizeof(Method));
	m->technique = TECHNIQUE_MPAD;
	
	m->Econd = EncodePotential(device, Econd);
	m->tCond = tCond;
	m->Edep = EncodePotential(device, Edep);
	m->tDep = tDep;
	m->tEquil = tEquil;
	
	m->cr_min = (uint8_t)cr_min;
	m->cr_max = (uint8_t)cr_max;
	m->cr = (uint8_t)cr_max;
	
	m->Ebegin = EncodePotential(device, E1);
	m->E1 = EncodePotential(device, E1);
	m->E2 = EncodePotential(device, E2);
	m->E3 = EncodePotential(device, E3);

	m->t2 = round(t2 / 71.5e-6);
	m->t3 = round(t3 / 71.5e-6);
	
	double tInt = t1 + t2 + t3;
	m->nPoints = floor((tRun / (tInt)) + 1);
	
	m->options = (uint16_t) options;
	m->ChAux = (uint8_t) CHAN_AUX_INPUT;
	m->Estby = EncodePotential(device, Estby);
	
	SetInterval(m, device, tInt);
	SetADTiming(m, device, t1 / 2, power_freq);
}

double CurrentToCR(double current_uA, CurrentRange cr)
{
	// CR 0 is nA, we want uA, so subtract 3
	return current_uA / pow(10.0, (double)cr - 3.0);
}

double CurrentTo_uA(double currentInCR, CurrentRange cr)
{
	// CR 0 is nA, we want uA, so subtract 3
	return currentInCR * pow(10.0, (double)cr - 3.0);
}

const char* GetCRString(CurrentRange cr)
{
	switch(cr)
	{
		case CR_1nA:
			return "1 nA";
		case CR_10nA:
			return "10 nA";
		case CR_100nA:
			return "100 nA";
		case CR_1uA:
			return "1 uA";
		case CR_10uA:
			return "10 uA";
		case CR_100uA:
			return "100 uA";
		case CR_1mA:
			return "1 mA";
		case CR_10mA:
			return "10 mA";
		case CR_100mA:
			return "100 mA";
	}
	return "?";
}


uint16_t EncodePotential(DeviceType device, double potential)
{
	return round((potential / EmstatVals[device][DACFACTOR] + 2.048) * 16000);
}

uint16_t EncodePotentialStep(DeviceType device, double potential_step)
{
	int16_t IntValue = round(potential_step / EmstatVals[device][DACFACTOR] * 16000);
	if (IntValue < 0)
	{
		IntValue += UINT16_MAX + 1;
	}
	return (uint16_t)IntValue;
}

CurrentRange CRFromCurrent(DeviceType device, double current_uA)
{
	return (CurrentRange)(log (current_uA) / log (10) + 3.5f);
}

double DecodePotential(DeviceType device, uint8_t lowByte, uint8_t highByte, DeviceParams param)
{
	return ((double)((((uint16_t)highByte) << 8) + lowByte) / (double)0xFFFF * 4.096 - 2.048) * EmstatVals[device][(int)param]; 
}

double DecodeAuxPotential(DeviceType device, uint8_t lowByte, uint8_t highByte)
{
	return ((double)((((uint16_t)highByte) << 8) + lowByte) / (double)0xFFFF * 4.096);
}

double DecodeCurrent(DeviceType device, uint8_t lowByte, uint8_t highByte, 
	uint8_t correctionByte, CurrentRange crange, DeviceParams param)
{
	double current = ((double)((((uint16_t)highByte) << 8) + lowByte) / (double)0xFFFF * 4.096 - 2.048);
	if(correctionByte == 0x01)
	{
		current += current;
	} 
	else if(correctionByte == 0xFF)
	{
		current -= current;
	}
	
	return CurrentTo_uA(current, crange);
}

double DecodeNoise(DeviceType device, uint8_t lowByte, uint8_t highByte)
{
	return (double)(lowByte + (((uint16_t)highByte)  << 8) ) * 1.1 / 1000.0 / 16.0f;
}

double SetADTiming(Method* m, DeviceType device, double tmeas, uint16_t power_freq)
{
	int8_t nadmean, d1, d16; // Values to calculate and send to EmStat
	// that define current sampling window
	double ADT16ad;
	if (power_freq == 60) // Where PowerFreq is the mains frequency in Hz
	{ 
		// Normally 50 Hz in Europe
		ADT16ad = 0.0002604;
		d1 = 5;
		d16 = 1;
	}
	else
	{ 
		// Power frequency = 50
		ADT16ad = 0.0003125;
		d1 = 11;
		d16 = 14;
	}
	if (tmeas < (1.0 / power_freq)) // The sampling time is shorter than
	{ 								// PowerLinePeriod (1/PowerFreq)
		ADT16ad = 0.000222; 		// So PowerLinePeriod=0.020 for 50 Hz
		d1 = 0; d16 = 0;
	}
	int32_t ncycles = (int32_t)(tmeas / ADT16ad);
	if (ncycles < 1)
		ncycles = 1; 	// Prevent errors
	
	//Where Math.Log calculates natural log:
	nadmean = round(log(ncycles) / log(2.0));
	
	if (nadmean < 0)
		nadmean = 0;
	if (nadmean > 11)
		nadmean = 11;
	
	m->nadmean = nadmean;
	m->d1 = d1;
	m->d16 = d16;
	
	// the following value can be used to verify the result:
	return (double)(pow(2.0, nadmean)) * ADT16ad;
	
	// so tmeas is now the actual sampling time based on the device resolution.
}

RetCode SetInterval(Method* m, DeviceType device, double interval)
{
	uint16_t T2HL = 0;
	uint16_t T2M = 0;
	uint32_t value = 0;
	double T2Value;
	double core_freq = 16.7772E+06; // clock frequency
	if (interval < 0.98) // maximum interval time is 255 * 0xFFFF / 16.7772E6 = 0.996 s
	{
		T2Value = interval * core_freq;
		T2M = round(T2Value / 0xFFFF + 1);
		// set T2 to highest possible values, in order to reduce ISR rate.
		if (T2M > 255)
		{
			T2M = 255;
		}
		if (T2M < 1)
		{
			T2M = 1;
		}
		T2HL = round(0xFFFF - T2Value / T2M);
		value = (uint32_t)4 << 24;
		value += (uint32_t)(T2M * 0xFFFF);
		value += T2HL;
	}
	else
	{
		if (interval / (60 * 60) >= 0x100)
		{
			return CODE_OUT_OF_RANGE;
		}
		else if (interval / 60 >= 0x100)
		{
			value = (uint32_t)3 << 24; //Hours
			value += round(interval / (60 * 60));
		}
		else if (interval >= 0x100)
		{
			value = (uint32_t)2 << 24; //Minutes
			value += round(interval / 60);
		}
		else if (interval * 128 >= 0x100)
		{
			value = (uint32_t)1 << 24; //Seconds
			value += round(interval);
		}
		else
		{
			value = 0; //128th seconds
			value += round(interval * 128);
		}
	}
	m->tInt = value;
	
	return CODE_OK;
}