/* ----------------------------------------------------------------------------
 *         PalmSens EmStat SDK
 * ----------------------------------------------------------------------------
 * Copyright (c) 2016, PalmSens BV
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * PalmSens's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY PALMSENS "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL PALMSENS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */

/**
 *  PSMethod encapsulates the format a measurement is sent in.
 *	For most scenarios, just create a Method struct and call the appropriate 
 *	SetParams function on it.
 *
 *	For more information about each technique (measurement type), see PSTrace.
 *	To get to this information, select a technique and click the question mark icon next to it.
 *
 *  Author(s): Hielke Veringa (hielke@palmsens.com)
 * 	Version: 1.23
 */
 
#ifndef PSMETHOD_H
#define PSMETHOD_H

//////////////////////////////////////////////////////////////////////////////
// Includes
//////////////////////////////////////////////////////////////////////////////

#include <stdint.h>
#include <stdbool.h>
#include "PSCommon.h"

//////////////////////////////////////////////////////////////////////////////
// Defines
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// Types
//////////////////////////////////////////////////////////////////////////////

///
/// The device types supported by this SDK.
///
typedef enum _DeviceType
{
	EMSTAT2,
	EMSTAT3,
	EMSTAT3P,
	NUM_DEVICES
} DeviceType;

///
/// Device specific parameter definitions
///
typedef enum _DeviceParams
{
	EFACTOR,
	DACFACTOR,
	VRANGE,
	NUM_DEVICE_PARAMS
} DeviceParams;

///
/// CON2 analog input channel
///
typedef enum _Channel
{
	CHAN_CURRENT		= 0x00,
	CHAN_POTENTIAL		= 0x01,
	CHAN_AUX_INPUT		= 0x02,
} Channel;

///
/// EmStat's possible current range settings.
///
typedef enum _CurrentRange 
{
  	CR_1nA 			= 0,
  	CR_10nA 		= 1,
  	CR_100nA 		= 2,
  	CR_1uA 			= 3,
  	CR_10uA 		= 4,
  	CR_100uA 		= 5,
  	CR_1mA 			= 6,
  	CR_10mA 		= 7,
	CR_100mA		= 8	//ES3+ new
} CurrentRange;

///
/// EmStat supported potentiostat techniques.
///
typedef enum _Technique 
{
	TECHNIQUE_LSV 		= 0,	/* Linear sweep */
	TECHNIQUE_DPV 		= 1,	/* Differential Pulse Voltammetry */
	TECHNIQUE_SWV 		= 2,  	/* Square Wave Voltammetry */
	TECHNIQUE_NPV 		= 3,	/* Normal Pulse Voltammetry */
	TECHNIQUE_CV 		= 5,	/* Cyclic Voltammetry */
	TECHNIQUE_AD		= 7,	/* Amperometric Detection */
	TECHNIQUE_PAD		= 8,	/* Pulsed Amperometric Detection */
	TECHNIQUE_FCA		= 9,	/* Fast Amperometry */
	TECHNIQUE_CP		= 10,	/* Chrono Potentiometry */
	TECHNIQUE_MPAD		= 11,	/* Multiple Pulsed Amperometric Detection */	
} Technique;

///
/// Measurement options enum.
///
typedef enum _Options 
{
	OPTIONS_NONE = 0x0,
	OPTIONS_CELL_ON_AFTER_MEASUREMENT = 0x4,
	OPTIONS_USE_STIRRER = 0x8,
	OPTIONS_MEASURE_AUX_INPUT = 0x40,
	OPTIONS_MEASURE_I_FORWARD_REVERSE = 0x400,
} Options;

///
/// Pulsed Amperometric Detection modes
///
typedef enum _PADMode
{
	/// <
	/// Pulsed Amperometric Detection DC mode
	/// 
	PAD_MODES_DC = 0x0,
	/// 
	/// Pulsed Amperometric Detection Pulse mode
	/// 
	PAD_MODES_PULSE = 0x10,
	/// 
	/// Pulsed Amperometric Detection Differential mode
	/// 
	PAD_MODES_DIFFERENTIAL = 0x20,
} PADMode;

///
/// Method parameters. Populate this using the "SetParams" functions.
/// Read the "EmStat communication protocol" document if you want to set the parameters manually.
///
typedef struct _Method
{
	/* Technique number */
	Technique technique;

	/* Options blob */
	uint16_t options;

	/* Pretreatment */
	uint16_t Econd, Edep;
	uint16_t tCond, tDep, tEquil;

	/* Autoranging settings */
	uint8_t cr_min, cr_max;

	/* Method-specific data */
	uint16_t Ebegin;
	uint32_t tInt;

	uint16_t mux_delay;
	uint16_t nPoints;
	uint16_t Evtx1;
	uint16_t Evtx2;
	uint8_t nScans;
	uint16_t limitMax;
	uint16_t limitMin;
	uint8_t crLimitMin;
	uint8_t crLimitMax;
	uint16_t tPulse;
	uint16_t Epulse;
	uint16_t Estep;

	uint16_t E1;
	uint16_t E2;
	uint16_t E3;
	
	uint16_t t2;
	uint16_t t3;
	
	uint16_t current; //NOT USED, BUT FOR BACKWARDS COMPATIBILITY
	uint16_t Eend;
    uint8_t cr;
    uint8_t tMax;
    
    uint16_t Estby;
    uint8_t nmux;
	uint8_t ChAux;
	
	//unsigned char channels;
	
	/* AD settings */
	unsigned char nadmean, d1, d16;
} Method;

//////////////////////////////////////////////////////////////////////////////
// Constants
//////////////////////////////////////////////////////////////////////////////

extern const double EmstatVals[NUM_DEVICES][NUM_DEVICE_PARAMS];

//////////////////////////////////////////////////////////////////////////////
// Pre-made Method Functions
// 
// These functions allow you to easily populate a method struct without having to learn
// the ins and outs of the EmStat method struct.
//////////////////////////////////////////////////////////////////////////////
/// 
/// Populates a method struct with params for a "Linear Sweep Voltammetry" measurement.
/// For a detailed explanation of this technique, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options:	Method option bits, valid options are: 
///				OPTIONS_CELL_ON_AFTER_MEASUREMENT, OPTIONS_USE_STIRRER and 
///				OPTIONS_MEASURE_AUX_INPUT.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///
void SetLinearSweepParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Eend, double Estep, double scan_rate, 
	Options options, double Estby, uint8_t power_freq);
	
/// 
/// Populates a method struct with params for a "Differential Pulse Voltammetry" measurement.
/// For a detailed explanation of this technique, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options: 	Method option bits, valid options are: 
///				OPTIONS_CELL_ON_AFTER_MEASUREMENT and OPTIONS_USE_STIRRER.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///
void SetDifferentialPulseParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Eend, double Estep, double EPulse, double tPulse, double scan_rate, 
	Options options, double Estby, uint8_t power_freq);
	
/// 
/// Populates a method struct with params for a "Square Wave Voltammetry" measurement.
/// For a detailed explanation of this technique, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options:	Method option bits, valid options are: 
///				OPTIONS_CELL_ON_AFTER_MEASUREMENT, OPTIONS_USE_STIRRER and 
///				OPTIONS_MEASURE_I_FORWARD_REVERSE.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///
void SetSquareWaveParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Eend, double Estep, double amplitude, double frequency,
	Options options, double Estby, uint8_t power_freq);
	
/// 
/// Populates a method struct with params for a "Normal Pulse Voltammetry" measurement.
/// For a detailed explanation of this measurement type, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options: 	Method option bits, valid options are: 
///				OPTIONS_CELL_ON_AFTER_MEASUREMENT and OPTIONS_USE_STIRRER.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///
void SetNormalPulseParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Eend, double Estep, double tPulse, double scan_rate,
	Options options, double Estby, uint8_t power_freq);

/// 
/// Populates a method struct with params for a "Cyclic Voltammetry" measurement.
/// For a detailed explanation of this technique, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options: 	Method option bits, valid options are: 
///				OPTIONS_CELL_ON_AFTER_MEASUREMENT, OPTIONS_USE_STIRRER and 
///				OPTIONS_MEASURE_AUX_INPUT.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///
void SetCyclicVoltammetryParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Ebegin, double Evtx1, double Evtx2, double Estep, double scan_rate, uint8_t nScans,
	Options options, double Estby, uint8_t power_freq);

/// 
/// Populates a method struct with params for a "Amperometric Detection" measurement.
/// For a detailed explanation of this technique, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options:	Method option bits, valid options are: 
///				OPTIONS_CELL_ON_AFTER_MEASUREMENT, OPTIONS_USE_STIRRER and 
///				OPTIONS_MEASURE_AUX_INPUT
///				Note: alternating mux mode (0x2) is also possible, but not implemented in this SDK.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///
void SetAmperometricDetectionParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Edc, double tInterval, double tRun,
	Options options, double Estby, uint8_t power_freq);
	
/// 
/// Populates a method struct with params for a "Pulsed Amperometric Detection" measurement.
/// For a detailed explanation of this technique, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options:	Method option bits, valid options are: 
///				OPTION_CELL_ON_AFTER_MEASUREMENT and OPTIONS_USE_STIRRER.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///
void SetPulsedAmperometricDetectionParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double Edc, double Epulse, double tPulse, PADMode mode, double tInterval, double tRun,
	Options options, double Estby, uint8_t power_freq);

/// 
/// Populates a method struct with params for a "Open Circuit Chrono Potentiometry" measurement.
/// For a detailed explanation of this technique, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options:	Method option bits, valid options are: 
///				OPTIONS_CELL_ON_AFTER_MEASUREMENT, OPTIONS_USE_STIRRER and 
///				OPTIONS_MEASURE_AUX_INPUT
///				Note: alternating mux mode (0x2) is also possible, but not implemented in this SDK.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///	
void SetOpenCircuitPotentiometryParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep,
	double tInterval, double tRun,
	Options options, double Estby, uint8_t power_freq);

/// 
/// Populates a method struct with params for a "Multiple Pulsed Amperometric Detection" 
/// measurement.
/// For a detailed explanation of this technique, check out the Help function of PSTrace.
/// 
///	m:			Pointer to the method struct that should be filled.
/// device:		The device type this method will be sent to.
/// options:	Method option bits, valid options are: 
///				OPTIONS_CELL_ON_AFTER_MEASUREMENT and OPTIONS_USE_STIRRER.
/// power_freq:	The frequency of the local power grid, usually 60Hz for South / North America
///				and 50Hz for the rest of the world.
///
void SetMultiplePulseAmperometryParams(Method* m, DeviceType device,
	double Econd, double tCond, double Edep, double tDep, 
	double tEquil, CurrentRange cr_min, CurrentRange cr_max, 
	double E1, double t1, double E2, double t2, double E3, double t3, double tRun,
	Options options, double Estby, uint8_t power_freq);
	
//////////////////////////////////////////////////////////////////////////////
// Util Functions
//////////////////////////////////////////////////////////////////////////////

///
/// Converts a current expressed in uA to a current expressed in the current range.
///
double CurrentToCR(double current_uA, CurrentRange cr);

///
/// Converts a current expressed in the current range to a current expressed in uA.
///
double CurrentTo_uA(double currentInCR, CurrentRange cr);
	
///
/// Returns the string representation of a CurrentRange.
///
const char* GetCRString(CurrentRange cr);

//////////////////////////////////////////////////////////////////////////////
// Helper Functions
//
// These functions are used internally by the SetParams functions.
// They can also be used to set method params manually.
//////////////////////////////////////////////////////////////////////////////
uint16_t EncodePotential(DeviceType device, double potential);
uint16_t EncodePotentialStep(DeviceType device, double potential_step);

//CurrentRange CRFromCurrent(DeviceType device, double current_uA);

double DecodePotential(DeviceType device, uint8_t lowByte, uint8_t highByte, DeviceParams param);
double DecodeCurrent(DeviceType device, uint8_t lowByte, uint8_t highByte, 
	uint8_t correctionByte, CurrentRange crange, DeviceParams param);
double DecodeAuxPotential(DeviceType device, uint8_t lowByte, uint8_t highByte);
double DecodeNoise(DeviceType device, uint8_t lowByte, uint8_t highByte);

double SetADTiming(Method* m, DeviceType device, double tmeas, uint16_t power_freq);
RetCode SetInterval(Method* m, DeviceType device, double interval);

#endif //PSMETHOD_H
